from __future__ import annotations

from abc import ABCMeta
from abc import abstractmethod
from functools import lru_cache
from functools import reduce
from functools import wraps
from operator import or_
from typing import Any
from typing import Dict
from typing import Generic
from typing import Hashable
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from pyworkflows.errors import UnhashableParameterError


T = TypeVar("T")


class TaskMeta(ABCMeta):
    def __new__(
        mcs: Type[TaskMeta],
        name: str,
        bases: Union[Type, Tuple[Type, ...]],
        namespace: Dict[str, Any],
    ) -> TaskMeta:
        init_str = "__init__"
        try:
            init = namespace[init_str]
        except KeyError:
            pass
        else:

            @wraps(init)
            def __init__(self: Task, *args: Any, **kwargs: Any) -> None:
                try:
                    self._dependencies_and_parameters
                except AttributeError:
                    pre_attrs = set(dir(self))
                    init(self, *args, **kwargs)
                    new_attrs: Dict[str, Any] = {
                        attr: getattr(self, attr)
                        for attr in set(dir(self)) - pre_attrs
                        if attr != "_dependencies_and_parameters"
                    }
                    dependencies: Set[Task] = {
                        x for x in new_attrs.values() if isinstance(x, Task)
                    }
                    parameters: Dict[str, Hashable] = {
                        k: self._check_hashable(k, v)
                        for k, v in new_attrs.items()
                        if not isinstance(v, Task)
                    }
                    self._dependencies_and_parameters = (
                        dependencies,
                        parameters,
                    )
                else:
                    init(self, *args, **kwargs)
                for key, value in self.get_parameters().items():
                    self._check_hashable(key, value)

            namespace[init_str] = __init__
        return super().__new__(mcs, name, bases, namespace)  # type: ignore


class Task(Generic[T], metaclass=TaskMeta):
    _dependencies_and_parameters: Tuple[Set[Task], Dict[str, Hashable]]

    @abstractmethod
    def __call__(self: Task[T]) -> T:
        raise NotImplementedError

    def __eq__(self: Task[T], other: Any) -> bool:
        if isinstance(other, Task):
            return (
                type(self) == type(other)
                and self.get_parameters() == other.get_parameters()
            )
        else:
            return NotImplemented

    def __hash__(self: Task[T]) -> int:
        return hash(frozenset(self.get_parameters().items()))

    @lru_cache
    def get_dependencies(self: Task[T], *, recurse: bool = False) -> Set[Task]:
        if recurse:
            dependencies = self.get_dependencies(recurse=False)
            while True:
                expanded = dependencies | set(
                    reduce(
                        or_,
                        (
                            x.get_dependencies(recurse=False)
                            for x in dependencies
                        ),
                        set(),
                    ),
                )
                if expanded.issubset(dependencies):
                    return dependencies
                else:
                    dependencies = dependencies | expanded
        else:
            try:
                dependencies, _ = self._dependencies_and_parameters
            except AttributeError:
                return set()
            else:
                return dependencies

    def get_parameters(self: Task[T]) -> Dict[str, Hashable]:
        try:
            _, parameters = self._dependencies_and_parameters
        except AttributeError:
            return {}
        else:
            return parameters

    @classmethod
    def _check_hashable(cls: Type[Task], key: str, value: Any) -> Hashable:
        try:
            hash(value)
        except TypeError:
            raise UnhashableParameterError(
                f"Parameter {cls.__name__}.{key}={value} is not hashable",
            ) from None
        else:
            return value



from __future__ import annotations

from pyworkflows.file_task import FileTask
from pyworkflows.file_task import floor_datetime
from pyworkflows.task import Task

__all__ = ["FileTask", "floor_datetime", "Task"]
__version__ = "0.2.4"



from __future__ import annotations

import datetime as dt
from abc import abstractmethod
from concurrent.futures._base import Future
from concurrent.futures.process import ProcessPoolExecutor
from dataclasses import dataclass
from functools import reduce
from itertools import count
from operator import or_
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Type
from typing import TypeVar
from typing import Union

from atomic_write_path import atomic_write_path
from dateutil.rrule import rrule
from dateutil.tz import EPOCH  # type: ignore
from dateutil.tz import gettz
from dateutil.tz import UTC

from pyworkflows.errors import BeforeTheEpochError
from pyworkflows.errors import RunTasksError
from pyworkflows.task import Task


T = TypeVar("T")
_DEFAULT_EXIST_OK = False
_DEFAULT_OVERWRITE = False
_DEFAULT_PARALLEL = False


class FileTask(Task[T]):
    def __call__(
        self: FileTask[T],
        *,
        overwrite: bool = _DEFAULT_OVERWRITE,
        exist_ok: bool = _DEFAULT_EXIST_OK,
    ) -> T:
        path = self.path()
        if overwrite:
            out = self.compute()
            with atomic_write_path(path, overwrite=True) as temp:
                self.write(out, temp)
            return out
        else:
            try:
                return self.read(path)
            except OSError:
                out = self.compute()
                try:
                    with atomic_write_path(path, overwrite=False) as temp:
                        self.write(out, temp)
                except FileExistsError:
                    if exist_ok:
                        return out
                    else:
                        raise
                else:
                    return out

    @abstractmethod
    def compute(self: FileTask[T]) -> T:
        raise NotImplementedError

    @abstractmethod
    def path(self: FileTask) -> Path:
        raise NotImplementedError

    @abstractmethod
    def read(self: FileTask[T], path: Path) -> T:
        raise NotImplementedError(path)

    @abstractmethod
    def write(self: FileTask, value: T, path: Path) -> None:
        raise NotImplementedError(value, path)

    @classmethod
    def run(
        cls: Type[FileTask],
        tasks: Iterable[FileTask],
        *,
        overwrite: bool = _DEFAULT_OVERWRITE,
        exist_ok: bool = _DEFAULT_EXIST_OK,
        parallel: bool = _DEFAULT_PARALLEL,
    ) -> None:
        as_set: Set[FileTask] = reduce(
            or_,
            ({task} | task.get_dependencies(recurse=True) for task in tasks),
            set(),
        )
        if not overwrite:
            as_set = {task for task in as_set if not task.path().exists()}
        if parallel:
            futures: Dict[Task, Future] = {}
            with ProcessPoolExecutor() as executor:
                while as_set:
                    completed: Set[Task] = set()
                    for task in as_set:
                        try:
                            future = futures[task]
                        except KeyError:
                            if all(
                                dep not in as_set
                                for dep in task.get_dependencies(recurse=True)
                            ) and (task not in futures):
                                futures[task] = executor.submit(
                                    task,
                                    overwrite=overwrite,
                                    exist_ok=exist_ok,
                                )
                        else:
                            if future.done():
                                maybe_exception = future.exception()
                                if maybe_exception is None:
                                    completed.add(task)
                                else:
                                    try:
                                        raise RunTasksError(
                                            f"Task {task} failed",
                                        )
                                    except RunTasksError as error:
                                        raise maybe_exception from error
                    as_set -= completed
        else:
            while as_set:
                task = next(iter(as_set))
                task(overwrite=overwrite, exist_ok=exist_ok)
                as_set.remove(task)


def floor_datetime(
    freq: int,
    *,
    as_of: Union[
        dt.datetime, Callable[..., dt.datetime],
    ] = lambda: dt.datetime.now(tz=UTC),
    interval: int = 1,
    wkst: Optional[int] = None,
    bysetpos: Optional[Union[int, Sequence[int]]] = None,
    bymonth: Optional[Union[int, Sequence[int]]] = None,
    bymonthday: Optional[Union[int, Sequence[int]]] = None,
    byyearday: Optional[Union[int, Sequence[int]]] = None,
    byeaster: Optional[Union[int, Sequence[int]]] = None,
    byweekno: Optional[Union[int, Sequence[int]]] = None,
    byweekday: Optional[Union[int, Sequence[int]]] = None,
    byhour: Optional[Union[int, Sequence[int]]] = None,
    byminute: Optional[Union[int, Sequence[int]]] = None,
    bysecond: Optional[Union[int, Sequence[int]]] = None,
    tz: Union[str, dt.tzinfo] = UTC,
) -> dt.datetime:
    if isinstance(as_of, dt.datetime):
        as_of_use = as_of
    else:
        as_of_use = as_of()
    epoch_utc = EPOCH.replace(tzinfo=UTC)
    if as_of_use < epoch_utc:
        raise BeforeTheEpochError(
            f"{as_of_use} is before the Epoch ({epoch_utc})",
        )

    def get_rrule(
        dtstart: dt.datetime,
        interval: int,
        *,
        count: Optional[int] = None,
        until: Optional[dt.datetime] = None,
        cache: bool = False,
    ) -> rrule:
        return rrule(
            freq,
            dtstart=dtstart,
            interval=interval,
            wkst=wkst,
            count=count,
            until=until,
            bysetpos=bysetpos,
            bymonth=bymonth,
            bymonthday=bymonthday,
            byyearday=byyearday,
            byeaster=byeaster,
            byweekno=byweekno,
            byweekday=byweekday,
            byhour=byhour,
            byminute=byminute,
            bysecond=bysecond,
            cache=cache,
        )

    @dataclass
    class Period:
        start: dt.datetime
        end: dt.datetime
        scale: int

    def get_period(dtstart: dt.datetime) -> Period:
        for scale in (2 ** i for i in count()):
            _, end = get_rrule(
                dtstart=dtstart, interval=scale * interval, count=2,
            )
            if dtstart <= as_of_use <= end:
                break
            else:
                dtstart = end
        return Period(dtstart, end, scale)

    def bisect_period(period: Period) -> dt.datetime:
        if period.scale == 1:
            r = get_rrule(dtstart=period.start, interval=interval, count=100)
            *_, end = r
            if period.start <= as_of_use <= end:
                return r.before(as_of_use, inc=True)
            else:
                return inner(period.start)
        else:
            half_scale, _ = divmod(period.scale, 2)
            _, mid, _ = get_rrule(
                dtstart=period.start, interval=half_scale * interval, count=3,
            )
            if period.start <= as_of_use <= mid:
                start, end = period.start, mid
            else:
                start, end = mid, period.end
            return bisect_period(Period(start, end, half_scale))

    def inner(start: dt.datetime) -> dt.datetime:
        period = get_period(start)
        result = bisect_period(period)
        if isinstance(tz, dt.tzinfo):
            tz_use: Optional[dt.tzinfo] = tz
        else:
            tz_use = gettz(tz)
        return result.astimezone(tz_use)

    return inner(epoch_utc)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from re import MULTILINE
from re import search
from typing import List

from setuptools import find_packages
from setuptools import setup


@dataclass
class MetaData:
    name: str
    version: str
    description: str
    long_description: str
    install_requires: List[str]


def _get_metadata() -> MetaData:
    with open("README.md") as file:
        long_description = file.read()
    header, description = long_description.splitlines()
    if match := search(r"^# ([\w-]+)$", header):
        name = match.group(1)
    else:
        raise ValueError("Unable to determine 'name' from README.md")
    path = (
        Path(__file__)
        .resolve()
        .parent.joinpath(name.replace("-", "_"), "__init__.py")
    )
    with open(path) as file:
        if (
            match := search(
                r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]$',
                file.read(),
                MULTILINE,
            )
        ) :
            version = match.group(1)
        else:
            raise ValueError(f"Unable to determine 'version' from {path}")
    with open("requirements/core.txt") as file:
        install_requires = file.read().strip().split("\n")
    return MetaData(
        name=name,
        version=version,
        description=description,
        long_description=long_description,
        install_requires=install_requires,
    )


_METADATA = _get_metadata()


setup(
    name=_METADATA.name,
    version=_METADATA.version,
    author="Bao Wei",
    author_email="baowei.ur521@gmail.com",
    license="MIT",
    description=_METADATA.description,
    long_description=_METADATA.long_description,
    long_description_content_type="text/markdown",
    install_requires=_METADATA.install_requires,
    python_requires=">=3.8",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
)
